# -*- coding: utf-8 -*-
'''
Some helper tools for pyCA testing.
'''

import logging
import pycurl

from importlib import reload  # noqa


# Raise log level above maximum to silence logging in tests.
# Then ensure that the log leven cannot be reset
logging.getLogger('').setLevel(logging.CRITICAL * 100)
logging.getLogger('').setLevel = lambda x: True


class ShouldFailException(pycurl.error):
    pass


def should_fail(*args, **kwargs):
    raise ShouldFailException(None, 0)


def __terminate():
    global _terminate
    _terminate -= 1
    return _terminate < 0


def terminate_fn(num):
    global _terminate
    _terminate = num
    return __terminate


class CurlMock():
    CAINFO = 0
    HTTPAUTH = 1
    HTTPAUTH_DIGEST = 2
    HTTPHEADER = 3
    HTTPPOST = 4
    HTTP_CODE = 5
    SSL_VERIFYHOST = 6
    SSL_VERIFYPEER = 7
    URL = 8
    USERPWD = 9
    WRITEFUNCTION = 10
    FAILONERROR = 11
    FOLLOWLOCATION = 12

    def setopt(self, *args):
        pass

    def perform(self):
        pass

    def getinfo(self, *args):
        return 200

    def close(self):
        pass



# -*- coding: utf-8 -*-
'''
pyCA tests for the REST api.
'''

import json
import os
import os.path
import tempfile
import unittest

from pyca import ui, config, db


class TestPycaRestInterface(unittest.TestCase):

    content_type = 'application/vnd.api+json'
    headers = {
        'Authorization': 'Basic YWRtaW46b3BlbmNhc3Q=',
        'Content-Type': content_type
    }

    def setUp(self):
        self.fd1, self.dbfile = tempfile.mkstemp()
        config.config()['agent']['database'] = 'sqlite:///' + self.dbfile
        db.init()

    def tearDown(self):
        os.close(self.fd1)
        os.remove(self.dbfile)

    def add_test_event(self):
        event = db.RecordedEvent()
        event.uid = '123'
        event.start = 1
        event.end = 2
        event.set_data('')
        event.status = db.Status.FINISHED_UPLOADING
        session = db.get_session()
        session.add(event)
        session.commit()
        return db.RecordedEvent(event)

    def test_mediatype_param(self):
        # JSONAPI must respond with 415 when mediatype parameters are present
        param_headers = self.headers.copy()
        param_headers['Content-Type'] = self.headers['Content-Type'] + ' a=b;'
        with ui.app.test_request_context(headers=param_headers, method='PUT'):
            assert ui.jsonapi.internal_state().status_code == 415
            assert ui.jsonapi.events().status_code == 415
            assert ui.jsonapi.event().status_code == 415
            assert ui.jsonapi.delete_event().status_code == 415
            assert ui.jsonapi.modify_event().status_code == 415

    def test_servicestatus(self):
        # Without authentication
        with ui.app.test_request_context():
            assert ui.jsonapi.internal_state().status_code == 401

        # With authentication
        with ui.app.test_request_context(headers=self.headers):
            response = ui.jsonapi.internal_state()
            assert response.headers['Content-Type'] == self.content_type
            assert response.status_code == 200
            meta = json.loads(response.data.decode('utf-8'))['meta']
            for service, status in meta['services'].items():
                assert hasattr(db.Service, service.upper())
                assert status == db.ServiceStatus.str(db.ServiceStatus.STOPPED)

    def test_events(self):
        # Without authentication
        with ui.app.test_request_context():
            assert ui.jsonapi.events().status_code == 401

        # With authentication
        with ui.app.test_request_context(headers=self.headers):
            response = ui.jsonapi.events()
            assert response.status_code == 200
            assert response.headers['Content-Type'] == self.content_type
            self.assertEqual(json.loads(
                response.data.decode('utf-8')), dict(data=[]))

        # With authentication and event
        event = self.add_test_event()
        with ui.app.test_request_context(headers=self.headers):
            response = ui.jsonapi.events()
            assert response.status_code == 200
            assert response.headers['Content-Type'] == self.content_type
            events = json.loads(response.data.decode('utf-8'))['data'][0]
            assert events.get('id') == event.uid

    def test_event(self):
        # Without authentication
        with ui.app.test_request_context():
            assert ui.jsonapi.event().status_code == 401

        # With authentication but invalid uid
        with ui.app.test_request_context(headers=self.headers):
            response = ui.jsonapi.event('123')
            assert response.status_code == 404
            assert response.headers['Content-Type'] == self.content_type
            error = json.loads(response.data.decode('utf-8'))['errors'][0]
            assert error == dict(title='No event with specified uid',
                                 status=404)

        # With authentication and valid uid
        event = self.add_test_event()
        with ui.app.test_request_context(headers=self.headers):
            response = ui.jsonapi.event(event.uid)
            assert response.status_code == 200
            assert response.headers['Content-Type'] == self.content_type
            events = json.loads(response.data.decode('utf-8'))['data'][0]
            assert events.get('id') == event.uid

    def test_delete_event(self):
        # Without authentication
        with ui.app.test_request_context():
            assert ui.jsonapi.delete_event().status_code == 401

        # With authentication but invalid uid
        with ui.app.test_request_context(headers=self.headers):
            response = ui.jsonapi.delete_event('')
            assert response.status_code == 404
            assert response.headers['Content-Type'] == self.content_type
            error = json.loads(response.data.decode('utf-8'))['errors'][0]
            assert error == dict(title='No event with specified uid',
                                 status=404)

        # With authentication and valid uid
        event = self.add_test_event()
        with ui.app.test_request_context(headers=self.headers):
            response = ui.jsonapi.delete_event(event.uid)
            assert response.status_code == 204
            assert response.headers['Content-Type'] == self.content_type
            assert not response.data

        # With hard deletion
        event = self.add_test_event()
        directory = event.directory()
        # create a stub recording directory
        if not os.path.exists(directory):
            os.makedirs(directory)
        with ui.app.test_request_context(headers=self.headers,
                                         query_string='hard=true'):
            response = ui.jsonapi.delete_event(event.uid)
            assert response.status_code == 204
            assert response.headers['Content-Type'] == self.content_type
            assert not response.data
            assert not os.path.exists(directory)

    def test_modify_event(self):
        # Without authentication
        with ui.app.test_request_context():
            assert ui.jsonapi.modify_event().status_code == 401

        args = dict(headers=self.headers)

        # With authentication but no or invalid data
        for data in (None,
                     '{"type":"event","id":0,"attributes":{"invalid":0}}',
                     '{"type":"event","id":0,"attributes":\
                     {"status":"invalid"}}'):
            args['data'] = {'data': [data]}
            with ui.app.test_request_context(**args):
                response = ui.jsonapi.modify_event(0)
                assert response.status_code == 400
                assert response.headers['Content-Type'] == self.content_type
                error = json.loads(response.data.decode('utf-8'))['errors'][0]
            assert error == dict(title='Invalid data', status=400)

        # With authentication but invalid uid
        content = {
            'data': [{
                'type': 'event',
                'id': 'invalid',
                'attributes': {
                    'start': 1000,
                    'end': 2000
                }
            }]
        }
        args['data'] = json.dumps(content)
        with ui.app.test_request_context(**args):
            response = ui.jsonapi.modify_event('')
            assert response.status_code == 400
            assert response.headers['Content-Type'] == self.content_type
            error = json.loads(response.data.decode('utf-8'))['errors'][0]
            assert error == dict(title='Invalid data', status=400)

        # With authentication and valid uid
        event = self.add_test_event()
        content['data'][0]['id'] = event.uid
        args['data'] = json.dumps(content)
        with ui.app.test_request_context(**args):
            response = ui.jsonapi.modify_event(event.uid)
            assert response.status_code == 200
            assert response.headers['Content-Type'] == self.content_type
            jsonevent = json.loads(response.data.decode('utf-8'))['data'][0]
            assert jsonevent.get('id') == event.uid
            assert jsonevent['attributes'].get('start') == 1000
            assert jsonevent['attributes'].get('end') == 2000



# -*- coding: utf-8 -*-
'''
    python-capture-agent
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: 2014-2017, Lars Kiesow <lkiesow@uos.de>
    :license: LGPL – see license.lgpl for more details.
'''

from pyca.utils import timestamp, try_mkdir, configure_service, terminate
from pyca.utils import set_service_status, set_service_status_immediate
from pyca.utils import recording_state, update_event_status
from pyca.config import config
from pyca.db import get_session, RecordedEvent, UpcomingEvent, Status,\
                    Service, ServiceStatus
import glob
import logging
import os
import os.path
import sdnotify
import shlex
import signal
import subprocess
import sys
import time
import traceback

logger = logging.getLogger(__name__)
notify = sdnotify.SystemdNotifier()
captureproc = None


def sigterm_handler(signum, frame):
    '''Intercept sigterm and terminate all processes.
    '''
    if captureproc and captureproc.poll() is None:
        captureproc.terminate()
    terminate(True)
    sys.exit(0)


def start_capture(upcoming_event):
    '''Start the capture process, creating all necessary files and directories
    as well as ingesting the captured files if no backup mode is configured.
    '''
    logger.info('Start recording')

    # First move event to recording_event table
    db = get_session()
    event = db.query(RecordedEvent)\
              .filter(RecordedEvent.uid == upcoming_event.uid)\
              .filter(RecordedEvent.start == upcoming_event.start)\
              .first()
    if not event:
        event = RecordedEvent(upcoming_event)
        db.add(event)
        db.commit()

    try_mkdir(config('capture', 'directory'))
    try_mkdir(event.directory())

    # Set state
    update_event_status(event, Status.RECORDING)
    recording_state(event.uid, 'capturing')
    set_service_status_immediate(Service.CAPTURE, ServiceStatus.BUSY)

    # Recording
    files = recording_command(event)
    # [(flavor,path),…]
    event.set_tracks(list(zip(config('capture', 'flavors'), files)))
    db.commit()

    # Set status
    # If part files exist, its an partial recording
    p = any([glob.glob(f'{f}-part-*') for f in files])
    state = Status.PARTIAL_RECORDING if p else Status.FINISHED_RECORDING
    logger.info("Set %s to %s", event.uid, Status.str(state))
    update_event_status(event, state)
    recording_state(event.uid, 'capture_finished')
    set_service_status_immediate(Service.CAPTURE, ServiceStatus.IDLE)

    logger.info('Finished recording')


def safe_start_capture(event):
    '''Start a capture process but make sure to catch any errors during this
    process, log them but otherwise ignore them.
    '''
    try:
        start_capture(event)
    except Exception:
        logger.error('Recording failed')
        logger.error(traceback.format_exc())
        # Update state
        recording_state(event.uid, 'capture_error')
        update_event_status(event, Status.FAILED_RECORDING)
        set_service_status_immediate(Service.CAPTURE, ServiceStatus.IDLE)


def recording_command(event):
    '''Run the actual command to record the a/v material.
    '''
    conf = config('capture')
    # Prepare command line
    cmd = conf['command']
    cmd = cmd.replace('{{time}}', str(event.remaining_duration(timestamp())))
    cmd = cmd.replace('{{dir}}', event.directory())
    cmd = cmd.replace('{{name}}', event.name())
    cmd = cmd.replace('{{previewdir}}', conf['preview_dir'])

    # Parse files into list
    files = (f.replace('{{dir}}', event.directory()) for f in conf['files'])
    files = [f.replace('{{name}}', event.name()) for f in files]

    # Move existing files from previous failed recordings
    for f in files:
        if not os.path.exists(f):
            continue
        # New filename
        i = 0
        while True:
            new_filename = f'{f}-part-{i}'
            if not os.path.exists(new_filename):
                break
            i += 1
        # Move file
        os.rename(f, new_filename)
        logger.warning("Moved file %s to %s to keep it", f, new_filename)

    # Signal configuration
    sigterm_time = conf['sigterm_time']
    sigkill_time = conf['sigkill_time']
    sigcustom_time = conf['sigcustom_time']
    sigcustom_time = 0 if sigcustom_time < 0 else event.end + sigcustom_time
    sigterm_time = 0 if sigterm_time < 0 else event.end + sigterm_time
    sigkill_time = 0 if sigkill_time < 0 else event.end + sigkill_time

    # Launch capture command
    logger.info(cmd)
    args = shlex.split(cmd)
    DEVNULL = getattr(subprocess, 'DEVNULL', os.open(os.devnull, os.O_RDWR))
    captureproc = subprocess.Popen(args, stdin=DEVNULL)
    hasattr(subprocess, 'DEVNULL') or os.close(DEVNULL)

    # Set systemd status
    notify.notify('STATUS=Capturing')

    # Check process
    while captureproc.poll() is None:
        notify.notify('WATCHDOG=1')
        if sigcustom_time and timestamp() > sigcustom_time:
            logger.info("Sending custom signal to capture process")
            captureproc.send_signal(conf['sigcustom'])
            sigcustom_time = 0  # send only once
        if sigterm_time and timestamp() > sigterm_time:
            logger.info("Terminating capture process")
            captureproc.terminate()
            sigterm_time = 0  # send only once
        elif sigkill_time and timestamp() > sigkill_time:
            logger.warning("Killing capture process")
            captureproc.kill()
            sigkill_time = 0  # send only once
        time.sleep(0.1)

    # Remove preview files:
    for preview in conf['preview']:
        try:
            os.remove(preview.replace('{{previewdir}}', conf['preview_dir']))
        except OSError:
            logger.warning('Could not remove preview files')
            logger.warning(traceback.format_exc())

    # Check process for errors
    exitcode = conf['exit_code']
    if captureproc.poll() > 0 and captureproc.returncode != exitcode:
        raise RuntimeError('Recording failed (%i)' % captureproc.returncode)

    # Reset systemd status
    notify.notify('STATUS=Waiting')

    # files
    return files


def control_loop():
    '''Main loop of the capture agent, retrieving and checking the schedule as
    well as starting the capture process if necessry.
    '''
    set_service_status_immediate(Service.CAPTURE, ServiceStatus.IDLE)
    notify.notify('READY=1')
    notify.notify('STATUS=Waiting')
    while not terminate():
        notify.notify('WATCHDOG=1')
        # Get next recording
        event = get_session().query(UpcomingEvent)\
                             .filter(UpcomingEvent.start <= timestamp())\
                             .filter(UpcomingEvent.end > timestamp())\
                             .first()
        if event:
            safe_start_capture(event)
        time.sleep(1.0)
    logger.info('Shutting down capture service')
    set_service_status(Service.CAPTURE, ServiceStatus.STOPPED)


def run():
    '''Start the capture agent.
    '''
    signal.signal(signal.SIGTERM, sigterm_handler)
    configure_service('capture.admin')
    control_loop()



# -*- coding: utf-8 -*-
'''
Helper methods used for the web interface.
'''

from functools import wraps
from flask import request, Response
from pyca.config import config


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        auth = request.authorization
        if config('ui', 'password') and not auth \
                or auth.username != config('ui', 'username') \
                or auth.password != config('ui', 'password'):
            return Response('pyCA: Login required\n', 401,
                            {'WWW-Authenticate': 'Basic realm="pyCA Login"'})
        return f(*args, **kwargs)
    return decorated


def jsonapi_mediatype(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if request.headers.get('Content-Type') != 'application/vnd.api+json'\
                and request.method != 'GET':
            return Response('Unsupported Media Type '
                            '(expected: application/vnd.api+json)', 415)
        response = f(*args, **kwargs)
        response.headers['Content-Type'] = 'application/vnd.api+json'
        return response
    return decorated



# -*- coding: utf-8 -*-
'''
pyCA tests for schedule handling
'''

import os
import os.path
import tempfile
import unittest

from pyca import ui, config, db


class TestPycaUI(unittest.TestCase):

    auth = {'Authorization': 'Basic YWRtaW46b3BlbmNhc3Q='}

    def setUp(self):
        self.fd1, self.dbfile = tempfile.mkstemp()
        self.fd2, self.previewfile = tempfile.mkstemp()
        config.config()['capture']['preview'] = [self.previewfile]
        config.config()['agent']['database'] = 'sqlite:///' + self.dbfile
        db.init()

    def tearDown(self):
        os.close(self.fd1)
        os.close(self.fd2)
        os.remove(self.dbfile)
        os.remove(self.previewfile)

    def test_home(self):
        # Without authentication
        with ui.app.test_request_context():
            self.assertEqual(ui.home().status_code, 401)

        # With authentication
        with ui.app.test_request_context(headers=self.auth):
            self.assertEqual(ui.home().status_code, 302)

    def test_ui(self):
        # Without authentication
        with ui.app.test_request_context():
            assert ui.serve_image(0).status_code == 401

        # With authentication
        with ui.app.test_request_context(headers=self.auth):
            assert ui.serve_image(9)[1] == 404
        with ui.app.test_request_context(headers=self.auth):
            r = ui.serve_image(0)
            assert r.status_code == 200
            r.close()



# -*- coding: utf-8 -*-
'''
Default configuration for pyCA.
'''

import configobj
import logging
import logging.handlers
import os
import socket
import sys
from validate import Validator

logger = logging.getLogger(__name__)

__CFG = '''
[agent]
name             = string(default='')
update_frequency = integer(min=5, default=60)
cal_lookahead    = integer(min=0, default=14)
backup_mode      = boolean(default=false)
database         = string(default='sqlite:///pyca.db')

[capture]
directory        = string(default='./recordings')
command          = string(default='ffmpeg -nostats -re -f lavfi -r 25 -i testsrc -t {{time}} {{dir}}/{{name}}.webm')
flavors          = force_list(default=list('presenter/source'))
files            = force_list(default=list('{{dir}}/{{name}}.webm'))
preview_dir      = string(default='./recordings')
preview          = force_list(default=list())
sigcustom        = integer(min=1, default=2)
sigcustom_time   = integer(min=-1, default=-1)
sigterm_time     = integer(min=-1, default=-1)
sigkill_time     = integer(min=-1, default=120)
exit_code        = integer(min=0, max=255, default=0)

[server]
url              = string(default='https://develop.opencast.org')
auth_method      = option('basic', 'digest', default='digest')
username         = string(default='opencast_system_account')
password         = string(default='CHANGE_ME')
insecure         = boolean(default=False)
certificate      = string(default='')

[ui]
username         = string(default='admin')
password         = string(default='opencast')
refresh_rate     = integer(min=1, default=10)
url              = string(default='http://localhost:5000')

[logging]
syslog           = boolean(default=False)
stderr           = boolean(default=True)
file             = string(default='')
level            = option('debug', 'info', 'warning', 'error', default='info')
format           = string(default='[%(name)s:%(lineno)s:%(funcName)s()] [%(levelname)s] %(message)s')
'''  # noqa

cfgspec = __CFG.split('\n')

__config = None


def configuration_file(cfgfile):
    '''Find the best match for the configuration file.
    '''
    if cfgfile is not None:
        return cfgfile
    # If no file is explicitely specified, probe for the configuration file
    # location.
    cfg = './etc/pyca.conf'
    if not os.path.isfile(cfg):
        return '/etc/pyca/pyca.conf'
    return cfg


def update_configuration(cfgfile=None):
    '''Update configuration from file.

    :param cfgfile: Configuration file to load.
    '''
    configobj.DEFAULT_INTERPOLATION = 'template'
    cfgfile = configuration_file(cfgfile)
    cfg = configobj.ConfigObj(cfgfile, configspec=cfgspec, encoding='utf-8')
    validator = Validator()
    val = cfg.validate(validator)
    if val is not True:
        raise ValueError('Invalid configuration: %s' % val)
    if len(cfg['capture']['files']) != len(cfg['capture']['flavors']):
        raise ValueError('List of files and flavors do not match')
    if not cfg['agent']['name']:
        cfg['agent']['name'] = 'pyca@' + socket.gethostname()
    globals()['__config'] = cfg
    logger_init()
    if cfg['server'].get('url', '').endswith('/'):
        logger.warning('Base URL ends with /. This is most likely a '
                       'configuration error. The URL should contain nothing '
                       'of the service paths.')
    logger.info('Configuration loaded from %s', cfgfile)
    check()
    return cfg


def check():
    '''Check configuration for sanity.
    '''
    if config('server')['insecure']:
        logger.warning('HTTPS CHECKS ARE TURNED OFF. A SECURE CONNECTION IS '
                       'NOT GUARANTEED')
    if config('server')['certificate']:
        # Ensure certificate exists and is readable
        open(config('server')['certificate'], 'rb').close()
    if config('agent')['backup_mode']:
        logger.info('Agent runs in backup mode. No data will be sent to '
                    'Opencast')


def config(*args):
    '''Get a specific configuration value or the whole configuration, loading
    the configuration file if it was not before.

    :param key: optional configuration key to return
    :type key: string
    :return: dictionary containing the configuration or configuration value
    '''
    cfg = __config or update_configuration()
    for key in args:
        if cfg is None:
            return
        cfg = cfg.get(key)
    return cfg


def logger_init():
    '''Initialize logger based on configuration
    '''
    handlers = []
    logconf = config('logging')
    if logconf['syslog']:
        handlers.append(logging.handlers.SysLogHandler(address='/dev/log'))
    if logconf['stderr']:
        handlers.append(logging.StreamHandler(sys.stderr))
    if logconf['file']:
        handlers.append(logging.handlers.WatchedFileHandler(logconf['file']))
    for handler in handlers:
        handler.setFormatter(logging.Formatter(logconf['format']))
        logging.root.addHandler(handler)

    logging.root.setLevel(logconf['level'].upper())
    logger.info('Log level set to %s', logconf['level'])



# -*- coding: utf-8 -*-

'''
Tests for pyCA configuration
'''

import unittest

from pyca import config


class TestPycaConfig(unittest.TestCase):

    def test_check(self):
        config.config()['server']['insecure'] = True
        config.config()['server']['certificate'] = '/xxx'
        try:
            config.check()
            assert False
        except IOError:
            assert True



# -*- coding: utf-8 -*-
'''
    python-capture-agent
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: 2014-2017, Lars Kiesow <lkiesow@uos.de>
    :license: LGPL – see license.lgpl for more details.
'''

from pyca.config import config
from pyca import db
from datetime import datetime
from dateutil.tz import tzutc
import errno
import json
import logging
import os
import os.path
import pycurl
import time
from io import BytesIO as bio
from urllib.parse import quote as urlquote


logger = logging.getLogger(__name__)


def http_request(url, post_data=None):
    '''Make an HTTP request to a given URL with optional parameters.
    '''
    logger.debug('Requesting URL: %s', url)
    buf = bio()
    curl = pycurl.Curl()
    curl.setopt(curl.URL, url.encode('ascii', 'ignore'))

    # Disable HTTPS verification methods if insecure is set
    if config('server', 'insecure'):
        curl.setopt(curl.SSL_VERIFYPEER, 0)
        curl.setopt(curl.SSL_VERIFYHOST, 0)

    if config('server', 'certificate'):
        # Make sure verification methods are turned on
        curl.setopt(curl.SSL_VERIFYPEER, 1)
        curl.setopt(curl.SSL_VERIFYHOST, 2)
        # Import your certificates
        curl.setopt(pycurl.CAINFO, config('server', 'certificate'))

    if post_data:
        curl.setopt(curl.HTTPPOST, post_data)
    curl.setopt(curl.WRITEFUNCTION, buf.write)
    logger.debug('Using authentication method %s',
                 config('server')['auth_method'])
    if config('server')['auth_method'] == 'digest':
        curl.setopt(curl.HTTPHEADER, ['X-Requested-Auth: Digest'])
        curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_DIGEST)
    curl.setopt(pycurl.USERPWD, ':'.join([config('server', 'username'),
                                          config('server', 'password')]))
    curl.setopt(curl.FAILONERROR, True)
    curl.setopt(curl.FOLLOWLOCATION, True)
    curl.perform()
    curl.close()
    result = buf.getvalue()
    buf.close()
    return result


def get_service(service_type):
    '''Get available service endpoints for a given service type from the
    Opencast ServiceRegistry.
    '''
    endpoint = '/services/available.json?serviceType=' + str(service_type)
    url = config('server', 'url') + endpoint
    response = http_request(url).decode('utf-8')
    services = json.loads(response).get('services', {}).get('service', [])
    services = ensurelist(services)
    endpoints = [service['host'] + service['path'] for service in services
                 if service['online'] and service['active']]
    for endpoint in endpoints:
        logger.info(u'Endpoint for %s: %s', service_type, endpoint)
    return endpoints


def timestamp():
    '''Get current unix timestamp
    '''
    return int(datetime.now(tzutc()).timestamp())


def try_mkdir(directory):
    '''Try to create a directory. Pass without error if it already exists.
    '''
    try:
        os.mkdir(directory)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise err


def configure_service(service):
    '''Get the location of a given service from Opencast and add it to the
    current configuration.
    '''
    while not config('service-' + service) and not terminate():
        try:
            config()['service-' + service] = \
                get_service('org.opencastproject.' + service)
        except pycurl.error as e:
            logger.error('Could not get %s endpoint: %s. Retrying in 5s',
                         service, e)
            time.sleep(5.0)


def ensurelist(x):
    '''Ensure an element is a list.'''
    return x if type(x) == list else [x]


def register_ca(status='idle'):
    '''Register this capture agent at the Matterhorn admin server so that it
    shows up in the admin interface.

    :param address: Address of the capture agent web ui
    :param status: Current status of the capture agent
    '''
    # If this is a backup CA we don't tell the Matterhorn core that we are
    # here.  We will just run silently in the background:
    if config('agent', 'backup_mode'):
        return
    service_endpoint = config('service-capture.admin')
    if not service_endpoint:
        logger.warning('Missing endpoint for updating agent status.')
        return
    params = [('address', config('ui', 'url')), ('state', status)]
    name = urlquote(config('agent', 'name').encode('utf-8'), safe='')
    url = f'{service_endpoint[0]}/agents/{name}'
    try:
        response = http_request(url, params).decode('utf-8')
        if response:
            logger.info(response)
    except pycurl.error as e:
        logger.warning('Could not set agent state to %s: %s', status, e)


def recording_state(recording_id, status):
    '''Send the state of the current recording to the Matterhorn core.

    :param recording_id: ID of the current recording
    :param status: Status of the recording
    '''
    # If this is a backup CA we do not update the recording state since the
    # actual CA does that and we want to interfere.  We will just run silently
    # in the background:
    if config('agent', 'backup_mode'):
        return
    params = [('state', status)]
    url = config('service-capture.admin')[0]
    url += '/recordings/%s' % recording_id
    try:
        result = http_request(url, params).decode('utf-8')
        logger.info(result)
    except pycurl.error as e:
        logger.warning('Could not set recording state to %s: %s', status, e)


@db.with_session
def update_event_status(dbs, event, status):
    '''Update the status of a particular event in the database.
    '''
    dbs.query(db.RecordedEvent).filter(db.RecordedEvent.start == event.start)\
                               .update({'status': status})
    event.status = status
    dbs.commit()


@db.with_session
def set_service_status(dbs, service, status):
    '''Update the status of a particular service in the database.
    '''
    srv = db.ServiceStates()
    srv.type = service
    srv.status = status

    dbs.merge(srv)
    dbs.commit()


def set_service_status_immediate(service, status):
    '''Update the status of a particular service in the database and send an
    immediate signal to Opencast.
    '''
    set_service_status(service, status)
    update_agent_state()


@db.with_session
def get_service_status(dbs, service):
    '''Update the status of a particular service in the database.
    '''
    srvs = dbs.query(db.ServiceStates).filter(db.ServiceStates.type == service)

    if srvs.count():
        return srvs[0].status

    return db.ServiceStatus.STOPPED


def update_agent_state():
    '''Update the current agent state in opencast.
    '''
    configure_service('capture.admin')
    status = 'idle'

    # Determine reported agent state with priority list
    if get_service_status(db.Service.SCHEDULE) == db.ServiceStatus.STOPPED:
        status = 'offline'
    elif get_service_status(db.Service.CAPTURE) == db.ServiceStatus.BUSY:
        status = 'capturing'
    elif get_service_status(db.Service.INGEST) == db.ServiceStatus.BUSY:
        status = 'uploading'

    register_ca(status=status)


def terminate(shutdown=None):
    '''Mark process as to be terminated.
    '''
    global _terminate
    if shutdown is not None:
        _terminate = shutdown
    return '_terminate' in globals() and _terminate



# -*- coding: utf-8 -*-

'''
Tests for basic capturing
'''

import os
import tempfile
import unittest

from pyca import utils, config, db
from tests.tools import should_fail, CurlMock, reload


class TestPycaUtils(unittest.TestCase):

    def setUp(self):
        config.config()['service-capture.admin'] = ['']

        # db
        self.fd, self.dbfile = tempfile.mkstemp()
        config.config()['agent']['database'] = 'sqlite:///' + self.dbfile
        db.init()

    def tearDown(self):
        os.close(self.fd)
        os.remove(self.dbfile)
        reload(utils)
        reload(config)

    def test_get_service(self):
        res = '''{"services":{
                    "service":{
                        "type":"org.opencastproject.capture.admin",
                        "host":"https://octestallinone.virtuos.uos.de",
                        "path":"/capture-admin",
                        "active":true,
                        "online":true,
                        "maintenance":false,
                        "jobproducer":false,
                        "onlinefrom":"2016-11-20T02:01:03.525+01:00",
                        "service_state":"NORMAL",
                        "state_changed":"2016-11-20T02:01:03.525+01:00",
                        "error_state_trigger":0,
                        "warning_state_trigger":0}}}'''.encode('utf-8')
        # Mock http_request method
        utils.http_request = lambda x, y=False: res
        endpoint = u'https://octestallinone.virtuos.uos.de/capture-admin'
        assert utils.get_service('') == [endpoint]

    def test_ensurelist(self):
        assert utils.ensurelist(1) == [1]
        assert utils.ensurelist([1]) == [1]

    def test_configure_service(self):
        utils.terminate(False)
        utils.get_service = lambda x: 'x'
        utils.configure_service('x')
        assert config.config()['service-x'] == 'x'

    def test_http_request(self):
        config.config()['server']['insecure'] = True
        config.config()['server']['certificate'] = 'nowhere'
        try:
            utils.http_request('http://127.0.0.1:8', [('x', 'y')])
            assert False
        except Exception as e:
            assert e.args[0] == 7  # connection error

    def test_http_request_mocked_curl(self):
        config.config()['server']['insecure'] = True
        config.config()['server']['certificate'] = 'nowhere'
        utils.pycurl.Curl = CurlMock
        try:
            utils.http_request('http://127.0.0.1:8', [('x', 'y')])
            assert True
        except Exception:
            assert False
        reload(utils.pycurl)

    def test_register_ca(self):
        utils.http_request = lambda x, y=False: b'xxx'
        utils.register_ca()
        utils.http_request = should_fail
        utils.register_ca()
        config.config()['agent']['backup_mode'] = True
        utils.register_ca()

    def test_recording_state(self):
        utils.http_request = lambda x, y=False: b'xxx'
        config.config()['service-capture.admin'] = ['']
        utils.recording_state('123', 'recording')
        utils.http_request = should_fail
        utils.recording_state('123', 'recording')
        config.config()['agent']['backup_mode'] = True
        utils.recording_state('123', 'recording')

    def test_set_service_status_immediate(self):
        utils.http_request = lambda x, y=False: b''
        utils.set_service_status_immediate(db.Service.SCHEDULE,
                                           db.ServiceStatus.IDLE)
        utils.set_service_status_immediate(db.Service.INGEST,
                                           db.ServiceStatus.BUSY)
        utils.set_service_status_immediate(db.Service.CAPTURE,
                                           db.ServiceStatus.BUSY)



# -*- coding: utf-8 -*-
'''
    python-capture-agent
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: 2014-2017, Lars Kiesow <lkiesow@uos.de>
    :license: LGPL – see license.lgpl for more details.
'''

from pyca.utils import set_service_status, update_agent_state, timestamp
from pyca.utils import terminate
from pyca.config import config
from pyca.db import Service, ServiceStatus
import logging
import sdnotify
import time

logger = logging.getLogger(__name__)
notify = sdnotify.SystemdNotifier()


def control_loop():
    '''Main loop, updating the capture agent state.
    '''
    set_service_status(Service.AGENTSTATE, ServiceStatus.BUSY)
    notify.notify('READY=1')
    notify.notify('STATUS=Running')
    while not terminate():
        notify.notify('WATCHDOG=1')

        next_update = timestamp() + config('agent', 'update_frequency')
        while not terminate() and timestamp() < next_update:
            time.sleep(0.1)

        if not terminate():
            update_agent_state()

    logger.info('Shutting down agentstate service')
    set_service_status(Service.AGENTSTATE, ServiceStatus.STOPPED)


def run():
    '''Start the capture agent state process.
    '''
    control_loop()



# -*- coding: utf-8 -*-
'''
Simple UI telling about the current state of the capture agent.
'''
from flask import Flask, send_from_directory, redirect, url_for
from pyca.config import config
from pyca.ui.utils import requires_auth
import os.path

__base_dir__ = os.path.abspath(os.path.dirname(__file__))
app = Flask(
    __name__,
    template_folder=os.path.join(__base_dir__, 'templates'),
    static_folder=os.path.join(__base_dir__, 'static'))
import pyca.ui.jsonapi  # noqa


@app.route('/')
@requires_auth
def home():
    '''Serve the status page of the capture agent.
    '''
    refresh_rate = config('ui', 'refresh_rate')

    return redirect(url_for(
        'static', filename='index.html', refresh=refresh_rate))


@app.route("/img/<int:image_id>")
@requires_auth
def serve_image(image_id):
    '''Serve the preview image with the given id
    '''
    try:
        preview_dir = config('capture', 'preview_dir')
        filepath = config('capture', 'preview')[image_id]
        filepath = filepath.replace('{{previewdir}}', preview_dir)
        filepath = os.path.abspath(filepath)
        if os.path.isfile(filepath):
            directory, filename = filepath.rsplit('/', 1)
            response = send_from_directory(directory, filename)
            response.cache_control.no_cache = True
            return response
    except (IndexError, KeyError):
        pass
    return '', 404



# -*- coding: utf-8 -*-
'''
    pyca.db
    ~~~~¨~~

    Database specification for pyCA
'''

import json
import os.path
import string
from pyca.config import config
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, Text, LargeBinary, DateTime, \
    create_engine
from sqlalchemy.orm import sessionmaker
from datetime import datetime
from functools import wraps
Base = declarative_base()


def init():
    '''Initialize connection to database. Additionally the basic database
    structure will be created if nonexistent.
    '''
    global engine
    engine = create_engine(config('agent', 'database'))
    Base.metadata.create_all(engine)


def get_session():
    '''Get a session for database communication. If necessary a new connection
    to the database will be established.

    :return:  Database session
    '''
    if 'engine' not in globals():
        init()
    Session = sessionmaker(bind=engine)
    return Session()


def with_session(f):
    """Wrapper for f to make a SQLAlchemy session present within the function

    :param f: Function to call
    :type f: Function
    :raises e: Possible exception of f
    :return: Result of f
    """
    @wraps(f)
    def decorated(*args, **kwargs):
        session = get_session()
        try:
            result = f(session, *args, **kwargs)
        except Exception as e:
            session.rollback()
            raise e
        finally:
            session.close()
        return result
    return decorated


class Constants():

    @classmethod
    def str(cls, value):
        '''Convert status (id) to its string name.'''
        for k, v in cls.__dict__.items():
            if k[0] in string.ascii_uppercase and v == value:
                return k.lower().replace('_', ' ')


class Status(Constants):
    '''Event status definitions
    '''
    UPCOMING = 1
    RECORDING = 2
    FAILED_RECORDING = 3
    FINISHED_RECORDING = 4
    UPLOADING = 5
    FAILED_UPLOADING = 6
    FINISHED_UPLOADING = 7
    PARTIAL_RECORDING = 8


class ServiceStatus(Constants):
    '''Service status type definitions
    '''
    STOPPED = 1
    IDLE = 2
    BUSY = 3


class Service(Constants):
    '''Service type definitions
    '''
    AGENTSTATE = 1
    CAPTURE = 2
    INGEST = 3
    SCHEDULE = 4


# Database Schema Definition
class BaseEvent():
    '''Database definition of an event.'''

    __tablename__ = 'event'

    uid = Column('uid', Text(), nullable=False, primary_key=True)
    start = Column('start', Integer(), primary_key=True)
    end = Column('end', Integer(), nullable=False)
    title = Column('title', Text())
    data = Column('data', LargeBinary(), nullable=False)
    status = Column('status', Integer(), nullable=False,
                    default=Status.UPCOMING)
    tracks = Column('tracks', LargeBinary(), nullable=True)

    def get_data(self):
        '''Load JSON data from event.
        '''
        return json.loads(self.data.decode('utf-8'))

    def set_data(self, data):
        '''Store data as JSON.
        '''
        # Python 3 wants bytes
        self.data = json.dumps(data).encode('utf-8')

    def name(self):
        '''Returns the filesystem name of this event.
        '''
        return 'recording-%i-%s' % (self.start, self.uid)

    def directory(self):
        '''Returns recording directory of this event.
        '''
        return os.path.join(config('capture', 'directory'), self.name())

    def remaining_duration(self, time):
        '''Returns the remaining duration for a recording.
        '''
        return max(0, self.end - max(self.start, time))

    def status_str(self):
        '''Return status as string.
        '''
        return Status.str(self.status)

    def get_tracks(self):
        '''Load JSON track data from event.
        '''
        if not self.tracks:
            return []
        return json.loads(self.tracks.decode('utf-8'))

    def set_tracks(self, tracks):
        '''Store track data as JSON.
        '''
        self.tracks = json.dumps(tracks).encode('utf-8')

    def __repr__(self):
        '''Return a string representation of an artist object.

        :return: String representation of object.
        '''
        return '<Event(start=%i, uid="%s")>' % (self.start, self.uid)

    def serialize(self):
        '''Serialize this object as dictionary usable for conversion to JSON.

        :return: Dictionary representing this object.
        '''
        return {
            'type': 'event',
            'id': self.uid,
            'attributes': {
                'start': self.start,
                'end': self.end,
                'uid': self.uid,
                'title': self.title,
                'data': self.get_data(),
                'status': Status.str(self.status)
            }
        }


class UpcomingEvent(Base, BaseEvent):
    '''List of upcoming events'''

    __tablename__ = 'upcoming_event'


class RecordedEvent(Base, BaseEvent):
    '''List of events pyca tried to record.'''

    __tablename__ = 'recorded_event'

    def __init__(self, event=None):
        if event:
            self.uid = event.uid
            self.start = event.start
            self.end = event.end
            self.title = event.title
            self.data = event.data
            self.status = event.status


class ServiceStates(Base):
    '''List of internal service states.'''

    __tablename__ = 'service_states'

    type = Column('type', Integer(), nullable=False, primary_key=True)
    status = Column('status', Integer(), nullable=False,
                    default=ServiceStatus.STOPPED)

    def __init__(self, service=None):
        if service:
            self.type = service.type
            self.status = service.status


class UpstreamState(Base):
    '''State of the upstream Opencast server.'''
    __tablename__ = 'upstream_state'
    url = Column('url', Text(), primary_key=True)
    last_synced = Column('last_synced', DateTime())

    @staticmethod
    def update_sync_time(url):
        s = get_session()
        s.merge(UpstreamState(url=url, last_synced=datetime.utcnow()))
        s.commit()
        s.close()



# -*- coding: utf-8 -*-
'''
pyCA tests for schedule handling
'''

import os
import sys
import tempfile
import unittest

from pyca import __main__, agentstate, capture, ingest, schedule, ui, utils
from tests.tools import should_fail, ShouldFailException, reload


class TestPycaMain(unittest.TestCase):

    def setUp(self):
        # Disable `print`
        # Ugly but compatible with Python 2
        sys.stdout = ShouldFailException()
        sys.stdout.write = lambda x: ''
        sys.stdout.flush = lambda: ''

    def teardown(self):
        reload(sys)

    def test_help(self):
        # test scenarios which end in the usage being printed
        sys.argv = ['pyca', '-h']
        try:
            __main__.main()
        except BaseException as e:
            assert e.code == 0
        sys.argv = ['pyca', '-x']
        try:
            __main__.main()
        except BaseException as e:
            assert e.code == 1
        sys.argv = ['pyca', 'too', 'many', 'arguments']
        try:
            __main__.main()
        except BaseException as e:
            assert e.code == 2
        sys.argv = ['pyca', 'fail']
        try:
            __main__.main()
        except BaseException as e:
            assert e.code == 3

    def test_broken_configuration_type(self):
        configs = ('[capture]\nflavors = "x/source"\nfiles = a, b',
                   '[agent]\nupdate_frequency = "nan"')
        for config in configs:
            fd, fn = tempfile.mkstemp()
            with open(fn, 'w') as f:
                f.write(config)
            sys.argv = ['pyca', '-c', fn, 'fail']
            try:
                __main__.main()
            except BaseException as e:
                assert e.code == 4
            os.close(fd)
            os.remove(fn)

    def test_run(self):
        for mod in (agentstate, capture, ingest, schedule):
            mod.run = should_fail
            sys.argv = ['pyca', mod.__name__.split('.')[-1]]
            try:
                __main__.main()
                assert False
            except ShouldFailException:
                assert True

        # Test ui start
        ui.app.run = should_fail
        sys.argv = ['pyca', 'ui']
        try:
            __main__.main()
            assert False
        except ShouldFailException:
            assert True

        # Test run all
        for mod in (agentstate, capture, ingest, schedule):
            mod.run = lambda: True
        sys.argv = ['pyca', 'run']
        try:
            __main__.main()
        except Exception:
            assert False

        for mod in (agentstate, capture, ingest, schedule):
            reload(mod)

    def test_sigterm(self):
        try:
            __main__.sigterm_handler(0, 0)
        except BaseException as e:
            assert e.code == 0
            assert utils.terminate()



# -*- coding: utf-8 -*-
'''
    python-capture-agent
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: 2014-2017, Lars Kiesow <lkiesow@uos.de>
    :license: LGPL – see license.lgpl for more details.
'''

import getopt
import multiprocessing
import signal
import sys
from pyca import capture, config, schedule, ingest, ui, agentstate, utils

USAGE = '''
Usage %s [OPTIONS] COMMAND

COMMANDS:
  run        --  Start all pyCA components except ui (default)
  capture    --  Start pyCA capture service
  ingest     --  Start pyCA ingest service
  schedule   --  Start pyCA schedule service
  agentstate --  Start pyCA agentstate service
  ui         --  Start web based user interface

OPTIONS:
  --help, -h             -- Show this help',
  --config=FILE, -c FILE -- Specify a config file',

CONFIGURATION:
  PyCA will try to find a configuration in the following order:
   - Configuration specified on the command line
   - ./etc/pyca.conf
   - /etc/pyca/pyca.conf
'''


def usage(retval=0):
    '''Print usage information to stdout and exit.
    '''
    print(USAGE % sys.argv[0])
    sys.exit(retval)


def sigint_handler(signum, frame):
    '''Intercept sigint and terminate services gracefully.
    '''
    utils.terminate(True)


def sigterm_handler(signum, frame):
    '''Intercept sigterm and terminate all processes.
    '''
    sigint_handler(signum, frame)
    for process in multiprocessing.active_children():
        process.terminate()
    sys.exit(0)


def run_all(*modules):
    '''Start all services.
    '''
    processes = [multiprocessing.Process(target=mod.run) for mod in modules]
    for p in processes:
        p.start()
    for p in processes:
        p.join()


def main():
    # Check command line options
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hc:', ['help', 'config='])
    except getopt.GetoptError:
        usage(1)

    cfg = None
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
        if opt in ('-c', '--config'):
            cfg = arg
            break

    # Make sure we got only one command
    if len(args) > 1:
        usage(2)

    try:
        config.update_configuration(cfg)
    except ValueError as e:
        print(str(e))
        sys.exit(4)

    # Set signal handler
    signal.signal(signal.SIGINT, sigint_handler)
    signal.signal(signal.SIGTERM, sigterm_handler)

    # Get the command with `run` as default
    cmd = (args + ['run'])[0]

    if cmd == 'run':
        run_all(schedule, capture, ingest, agentstate)
    elif cmd == 'schedule':
        schedule.run()
    elif cmd == 'capture':
        capture.run()
    elif cmd == 'ingest':
        ingest.run()
    elif cmd == 'agentstate':
        agentstate.run()
    elif cmd == 'ui':
        signal.signal(signal.SIGINT, signal.default_int_handler)
        ui.app.run(threaded=False)
    else:
        # Invalid command
        usage(3)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

'''
Tests for basic capturing
'''

import os
import os.path
import shutil
import tempfile
import unittest

from unittest.mock import patch

from pyca import ingest, config, db, utils
from tests.tools import should_fail, terminate_fn


class TestPycaIngest(unittest.TestCase):

    def setUp(self):
        ingest.http_request = lambda x, y=False: b'xxx'
        self.fd, self.dbfile = tempfile.mkstemp()
        self.cadir = tempfile.mkdtemp()
        config.config('agent')['database'] = 'sqlite:///' + self.dbfile
        config.config('capture')['directory'] = self.cadir
        config.config()['service-ingest'] = ['']
        config.config()['service-capture.admin'] = ['']

        # Mock event
        db.init()
        event = db.RecordedEvent()
        event.uid = '123123'
        event.status = db.Status.FINISHED_RECORDING
        event.start = utils.timestamp()
        event.end = event.start + 1
        prop = 'org.opencastproject.capture.agent.properties'
        dcns = 'http://www.opencastproject.org/xsd/1.0/dublincore/'
        data = [{'data': u'äü%sÄÜß' % dcns,
                 'fmttype': 'application/xml',
                 'x-apple-filename': 'episode.xml'},
                {'data': u'äü%sÄÜß' % dcns,
                 'fmttype': 'application/xml',
                 'x-apple-filename': 'series.xml'},
                {'data': u'event.title=äüÄÜß\n' +
                         u'org.opencastproject.workflow.config.x=123\n' +
                         u'org.opencastproject.workflow.definition=fast',
                 'fmttype': 'application/text',
                 'x-apple-filename': prop}]
        event.set_data({'attach': data})

        # Create recording
        os.mkdir(event.directory())
        trackfile = os.path.join(event.directory(), 'test.mp4')
        open(trackfile, 'wb').close()
        event.set_tracks([('presenter/source', trackfile)])
        session = db.get_session()
        session.add(event)
        session.commit()
        self.event = db.RecordedEvent(event)

    def tearDown(self):
        os.close(self.fd)
        os.remove(self.dbfile)
        shutil.rmtree(self.cadir)

    @patch(__name__+'.ingest.ingest')
    def test_safe_start_ingest(self, ingest_fn):
        ingest_fn.side_effect = lambda x: None
        ingest.safe_start_ingest(self.event)
        ingest_fn.side_effect = should_fail
        ingest.safe_start_ingest(self.event)

    def test_run(self):
        ingest.terminate(True)
        ingest.run()
        ingest.terminate = terminate_fn(1)
        ingest.run()
        config.config('agent')['backup_mode'] = True
        ingest.run()



# -*- coding: utf-8 -*-
'''
    python-capture-agent
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: 2014-2017, Lars Kiesow <lkiesow@uos.de>
    :license: LGPL – see license.lgpl for more details.
'''

from pyca.utils import http_request, configure_service, timestamp, \
                       set_service_status_immediate, terminate
from pyca.config import config
from pyca.db import get_session, UpcomingEvent, Service, ServiceStatus, \
    UpstreamState
from base64 import b64decode
from datetime import datetime
import dateutil.parser
import logging
import pycurl
import sdnotify
import time
import traceback
from urllib.parse import urlencode

logger = logging.getLogger(__name__)
notify = sdnotify.SystemdNotifier()


def parse_ical(vcal):
    '''Parse Opencast schedule iCalendar file and return events as dict
    '''
    vcal = vcal.replace('\r\n ', '').replace('\r\n\r\n', '\r\n')
    vevents = vcal.split('\r\nBEGIN:VEVENT\r\n')
    del(vevents[0])
    events = []
    for vevent in vevents:
        event = {}
        for line in vevent.split('\r\n'):
            line = line.split(':', 1)
            key = line[0].lower()
            if len(line) <= 1 or key == 'end':
                continue
            if key.startswith('dt'):
                event[key] = int(dateutil.parser.parse(line[1]).timestamp())
                continue
            if not key.startswith('attach'):
                event[key] = line[1]
                continue
            # finally handle attachments
            event['attach'] = event.get('attach', [])
            attachment = {}
            for x in [x.split('=') for x in line[0].split(';')]:
                if x[0].lower() in ['fmttype', 'x-apple-filename']:
                    attachment[x[0].lower()] = x[1]
            attachment['data'] = b64decode(line[1]).decode('utf-8')
            event['attach'].append(attachment)
        events.append(event)
    return events


def get_schedule():
    '''Try to load schedule from the Matterhorn core. Returns a valid schedule
    or None on failure.
    '''
    params = {'agentid': config('agent', 'name').encode('utf8')}
    lookahead = config('agent', 'cal_lookahead') * 24 * 60 * 60
    if lookahead:
        params['cutoff'] = str((timestamp() + lookahead) * 1000)
    uri = '%s/calendars?%s' % (config('service-scheduler')[0],
                               urlencode(params))
    try:
        vcal = http_request(uri)
        UpstreamState.update_sync_time(config('server', 'url'))
    except pycurl.error as e:
        logger.error('Could not get schedule: %s', e)
        return

    try:
        cal = parse_ical(vcal.decode('utf-8'))
    except Exception:
        logger.error('Could not parse ical')
        logger.error(traceback.format_exc())
        return
    db = get_session()
    db.query(UpcomingEvent).delete()
    for event in cal:
        # Ignore events that have already ended
        if event['dtend'] <= timestamp():
            continue
        e = UpcomingEvent()
        e.start = event['dtstart']
        e.end = event['dtend']
        e.uid = event.get('uid')
        e.title = event.get('summary')
        e.set_data(event)
        db.add(e)
    db.commit()


def control_loop():
    '''Main loop, retrieving the schedule.
    '''
    set_service_status_immediate(Service.SCHEDULE, ServiceStatus.BUSY)
    notify.notify('READY=1')
    while not terminate():
        notify.notify('WATCHDOG=1')
        # Try getting an updated schedule
        get_schedule()
        session = get_session()
        next_event = session.query(UpcomingEvent)\
                            .filter(UpcomingEvent.end > timestamp())\
                            .order_by(UpcomingEvent.start)\
                            .first()
        if next_event:
            logger.info('Next scheduled recording: %s',
                        datetime.fromtimestamp(next_event.start))
            notify.notify('STATUS=Next scheduled recording: %s' %
                          datetime.fromtimestamp(next_event.start))
        else:
            logger.info('No scheduled recording')
            notify.notify('STATUS=No scheduled recording')
        session.close()

        next_update = timestamp() + config('agent', 'update_frequency')
        while not terminate() and timestamp() < next_update:
            time.sleep(0.1)

    logger.info('Shutting down schedule service')
    set_service_status_immediate(Service.SCHEDULE, ServiceStatus.STOPPED)


def run():
    '''Start the capture agent.
    '''
    configure_service('scheduler')
    control_loop()



# -*- coding: utf-8 -*-

'''
Tests for basic capturing
'''

import os
import os.path
import shutil
import tempfile
import unittest

from pyca import capture, config, db, utils
from tests.tools import should_fail, terminate_fn, reload


class TestPycaCapture(unittest.TestCase):

    def setUp(self):
        utils.http_request = lambda x, y=False: b'xxx'
        self.fd, self.dbfile = tempfile.mkstemp()
        self.cadir = tempfile.mkdtemp()
        preview = os.path.join(self.cadir, 'preview.png')
        open(preview, 'a').close()
        config.config()['agent']['database'] = 'sqlite:///' + self.dbfile
        config.config()['capture']['command'] = 'touch {{dir}}/{{name}}.mp4'
        config.config()['capture']['directory'] = self.cadir
        config.config()['capture']['preview'] = [preview]
        config.config()['service-capture.admin'] = ['']

        # Mock event
        db.init()
        self.event = db.BaseEvent()
        self.event.uid = '123123'
        self.event.title = u'äüÄÜß'
        self.event.start = utils.timestamp()
        self.event.end = self.event.start
        self.event.status = db.Status.UPCOMING
        data = [{'data': u'äüÄÜß',
                 'fmttype': 'application/xml',
                 'x-apple-filename': 'episode.xml'},
                {'data': u'äüÄÜß',
                 'fmttype': 'application/xml',
                 'x-apple-filename': 'series.xml'},
                {'data': u'event.title=äüÄÜß\n' +
                         u'org.opencastproject.workflow.config.x=123\n' +
                         u'org.opencastproject.workflow.definition=fast',
                 'fmttype': 'application/text',
                 'x-apple-filename': 'org.opencastproject.capture.agent' +
                                     '.properties'}]
        self.event.set_data({'attach': data})

    def tearDown(self):
        os.close(self.fd)
        os.remove(self.dbfile)
        shutil.rmtree(self.cadir)
        reload(capture)
        reload(config)
        reload(utils)

    def test_start_capture(self):
        capture.start_capture(self.event)

    def test_start_capture_recording_command_failure(self):
        config.config()['capture']['command'] = 'false'
        try:
            capture.start_capture(self.event)
            assert False
        except RuntimeError:
            assert True

    def test_start_capture_sigterm(self):
        config.config()['capture']['command'] = 'sleep 10'
        config.config()['capture']['sigterm_time'] = 0
        capture.start_capture(self.event)

    def test_start_capture_sigkill(self):
        config.config()['capture']['command'] = 'sleep 10'
        config.config()['capture']['sigkill_time'] = 0
        capture.start_capture(self.event)

    def test_safe_start_capture(self):
        '''Ensure that safe_start_capture always returns without error to not
        disrupt the main loop.
        '''
        capture.start_capture = should_fail
        capture.safe_start_capture(self.event)

    def test_run(self):
        capture.terminate = terminate_fn(1)
        capture.run()

    def test_sigterm(self):
        try:
            capture.sigterm_handler(0, 0)
        except BaseException as e:
            assert e.code == 0
            assert utils.terminate()



from setuptools import setup, find_packages

setup(
    name="pyca",
    version="3.1",
    description="Opencast Capture Agent",
    author="Lars Kiesow",
    author_email='lkiesow@uos.de',
    license="LGPLv3",
    url="https://github.com/opencast/pyCA",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "pycurl>=7.19.5",
        "python-dateutil>=2.4.0",
        "configobj>=5.0.0",
        "sqlalchemy>=0.9.8",
        "sdnotify>=0.3.2",
        "psutil>=5.0.1",
        "flask>=1.0.2",
    ],
    entry_points={
        'console_scripts': [
            'pyca = pyca.__main__:main'
        ]
    },
    test_suite="tests",
)



# -*- coding: utf-8 -*-
'''
    python-capture-agent
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: 2014-2017, Lars Kiesow <lkiesow@uos.de>
    :license: LGPL – see license.lgpl for more details.
'''

from pyca.config import config
from pyca.db import get_session, RecordedEvent, Status, Service, ServiceStatus
from pyca.utils import http_request, configure_service, set_service_status
from pyca.utils import set_service_status_immediate, recording_state
from pyca.utils import update_event_status, terminate
from random import randrange
import logging
import pycurl
import sdnotify
import time
import traceback

logger = logging.getLogger(__name__)
notify = sdnotify.SystemdNotifier()


def get_config_params(properties):
    '''Extract the set of configuration parameters from the properties attached
    to the schedule
    '''
    param = []
    wdef = ''
    for prop in properties.split('\n'):
        if prop.startswith('org.opencastproject.workflow.config'):
            key, val = prop.split('=', 1)
            key = key.split('.')[-1]
            param.append((key, val))
        elif prop.startswith('org.opencastproject.workflow.definition'):
            wdef = prop.split('=', 1)[-1]
    return wdef, param


def ingest(event):
    '''Ingest a finished recording to the Opencast server.
    '''
    # Update status
    set_service_status(Service.INGEST, ServiceStatus.BUSY)
    notify.notify('STATUS=Uploading')
    recording_state(event.uid, 'uploading')
    update_event_status(event, Status.UPLOADING)

    # Select ingest service
    # The ingest service to use is selected at random from the available
    # ingest services to ensure that not every capture agent uses the same
    # service at the same time
    service = config('service-ingest')
    service = service[randrange(0, len(service))]
    logger.info('Selecting ingest service to use: ' + service)

    # create mediapackage
    logger.info('Creating new mediapackage')
    mediapackage = http_request(service + '/createMediaPackage')

    # extract workflow_def, workflow_config and add DC catalogs
    prop = 'org.opencastproject.capture.agent.properties'
    dcns = 'http://www.opencastproject.org/xsd/1.0/dublincore/'
    for attachment in event.get_data().get('attach'):
        data = attachment.get('data')
        if attachment.get('x-apple-filename') == prop:
            workflow_def, workflow_config = get_config_params(data)

        # Check for dublincore catalogs
        elif attachment.get('fmttype') == 'application/xml' and dcns in data:
            name = attachment.get('x-apple-filename', '').rsplit('.', 1)[0]
            logger.info('Adding %s DC catalog', name)
            fields = [('mediaPackage', mediapackage),
                      ('flavor', 'dublincore/%s' % name),
                      ('dublinCore', data.encode('utf-8'))]
            mediapackage = http_request(service + '/addDCCatalog', fields)

    # add track
    for (flavor, track) in event.get_tracks():
        logger.info('Adding track ({0} -> {1})'.format(flavor, track))
        track = track.encode('ascii', 'ignore')
        fields = [('mediaPackage', mediapackage), ('flavor', flavor),
                  ('BODY1', (pycurl.FORM_FILE, track))]
        mediapackage = http_request(service + '/addTrack', fields)

    # ingest
    logger.info('Ingest recording')
    fields = [('mediaPackage', mediapackage)]
    if workflow_def:
        fields.append(('workflowDefinitionId', workflow_def))
    if event.uid:
        fields.append(('workflowInstanceId',
                       event.uid.encode('ascii', 'ignore')))
    fields += workflow_config
    mediapackage = http_request(service + '/ingest', fields)

    # Update status
    recording_state(event.uid, 'upload_finished')
    update_event_status(event, Status.FINISHED_UPLOADING)
    notify.notify('STATUS=Running')
    set_service_status_immediate(Service.INGEST, ServiceStatus.IDLE)

    logger.info('Finished ingest')


def safe_start_ingest(event):
    '''Start a capture process but make sure to catch any errors during this
    process, log them but otherwise ignore them.
    '''
    try:
        ingest(event)
    except Exception:
        logger.error('Something went wrong during the upload')
        logger.error(traceback.format_exc())
        # Update state if something went wrong
        recording_state(event.uid, 'upload_error')
        update_event_status(event, Status.FAILED_UPLOADING)
        set_service_status_immediate(Service.INGEST, ServiceStatus.IDLE)


def control_loop():
    '''Main loop of the capture agent, retrieving and checking the schedule as
    well as starting the capture process if necessry.
    '''
    set_service_status_immediate(Service.INGEST, ServiceStatus.IDLE)
    notify.notify('READY=1')
    notify.notify('STATUS=Running')
    while not terminate():
        notify.notify('WATCHDOG=1')
        # Get next recording
        event = get_session().query(RecordedEvent)\
                             .filter(RecordedEvent.status ==
                                     Status.FINISHED_RECORDING).first()
        if event:
            safe_start_ingest(event)
        time.sleep(1.0)
    logger.info('Shutting down ingest service')
    set_service_status(Service.INGEST, ServiceStatus.STOPPED)


def run():
    '''Start the capture agent.
    '''
    # If we are a backup CA, we don't want to actually upload anything. So
    # let's just quit here and do not run the ingest service at all.
    if config('agent')['backup_mode']:
        return

    configure_service('ingest')
    configure_service('capture.admin')
    control_loop()



# -*- coding: utf-8 -*-
'''
pyCA tests for schedule handling
'''

import datetime
import os
import os.path
import tempfile
import unittest

from pyca import schedule, config, db, utils
from tests.tools import should_fail, ShouldFailException, terminate_fn, reload


class TestPycaCapture(unittest.TestCase):

    END = (datetime.datetime.today() + datetime.timedelta(days=1)).isoformat()

    VCAL = ('''BEGIN:VCALENDAR
        BEGIN:VEVENT
        UID:20170223T171244Z
        SUMMARY:TEST
        DTSTART:20170223T230000Z
        DTEND:%sZ
        ATTACH;FMTTYPE=application/xml;VALUE=BINARY;ENCODING=BASE64;
          X-APPLE-FILENAME=episode.xml:Li4u
        END:VEVENT
        END:VCALENDAR''' % END).replace('\n        ', '\r\n').encode('utf-8')

    def setUp(self):
        utils.http_request = lambda x, y=False: b'xxx'
        self.fd, self.dbfile = tempfile.mkstemp()
        config.config()['agent']['database'] = 'sqlite:///' + self.dbfile
        config.config()['service-scheduler'] = ['']
        config.config()['service-capture.admin'] = ['']

        # Mock event
        db.init()

    def tearDown(self):
        os.close(self.fd)
        os.remove(self.dbfile)
        reload(utils)
        reload(schedule)

    def test_get_schedule(self):
        # Failed request
        schedule.http_request = should_fail
        schedule.get_schedule()
        assert not db.get_session().query(db.UpcomingEvent).count()

        # Failed parsing ical
        schedule.http_request = lambda x: ShouldFailException
        schedule.get_schedule()
        assert not db.get_session().query(db.UpcomingEvent).count()

        # Get schedule
        schedule.http_request = lambda x: self.VCAL
        schedule.get_schedule()
        assert db.get_session().query(db.UpcomingEvent).count()

    def test_run(self):
        schedule.terminate = terminate_fn(2)
        schedule.run()



# -*- coding: utf-8 -*-
from flask import jsonify, make_response, request
from pyca.config import config
from pyca.db import Service, ServiceStatus, UpcomingEvent, \
    RecordedEvent, UpstreamState
from pyca.db import with_session, Status, ServiceStates
from pyca.ui import app
from pyca.ui.utils import requires_auth, jsonapi_mediatype
from pyca.utils import get_service_status, ensurelist
import logging
import os
import psutil
import shutil

logger = logging.getLogger(__name__)


def make_error_response(error, status=500):
    ''' Return a response with a jsonapi error object
    '''
    content = {
        'errors': [{
            'status': status,
            'title': error
        }]
    }
    return make_response(jsonify(content), status)


def make_data_response(data, status=200):
    ''' Return a response with a list of jsonapi data objects
    '''
    content = {'data': ensurelist(data)}
    return make_response(jsonify(content), status)


@app.route('/api/name')
@requires_auth
@jsonapi_mediatype
def get_name():
    '''Serve the name of the capure agent via json.
    '''
    return make_response(
        jsonify({'meta': {'name': config('agent', 'name')}}))


@app.route('/api/previews')
@requires_auth
@jsonapi_mediatype
def get_images():
    '''Serve the list of preview images via json.
    '''
    # Get IDs of existing preview images
    preview = config('capture', 'preview')
    previewdir = config('capture', 'preview_dir')
    preview = [p.replace('{{previewdir}}', previewdir) for p in preview]
    preview = zip(preview, range(len(preview)))

    # Create return
    preview = [{'attributes': {'id': p[1]}, 'id': str(p[1]), 'type': 'preview'}
               for p in preview if os.path.isfile(p[0])]
    return make_data_response(preview)


@app.route('/api/services')
@requires_auth
@jsonapi_mediatype
def internal_state():
    '''Serve a json representation of internal agentstate as meta data
    '''
    data = {'services': {
        'capture': ServiceStatus.str(get_service_status(Service.CAPTURE)),
        'ingest': ServiceStatus.str(get_service_status(Service.INGEST)),
        'schedule': ServiceStatus.str(get_service_status(Service.SCHEDULE)),
        'agentstate': ServiceStatus.str(get_service_status(Service.AGENTSTATE))
    }}
    return make_response(jsonify({'meta': data}))


@app.route('/api/events/')
@requires_auth
@jsonapi_mediatype
@with_session
def events(db):
    '''Serve a JSON representation of events splitted by upcoming and already
    recorded events.
    '''
    upcoming_events = db.query(UpcomingEvent)\
                        .order_by(UpcomingEvent.start)
    recorded_events = db.query(RecordedEvent)\
                        .order_by(RecordedEvent.start.desc())

    result = [event.serialize() for event in upcoming_events]
    result += [event.serialize() for event in recorded_events]
    return make_data_response(result)


@app.route('/api/events/<uid>')
@requires_auth
@jsonapi_mediatype
@with_session
def event(db, uid):
    '''Return a specific events JSON
    '''
    event = db.query(RecordedEvent).filter(RecordedEvent.uid == uid).first() \
        or db.query(UpcomingEvent).filter(UpcomingEvent.uid == uid).first()

    if event:
        return make_data_response(event.serialize())
    return make_error_response('No event with specified uid', 404)


@app.route('/api/events/<uid>', methods=['DELETE'])
@requires_auth
@jsonapi_mediatype
@with_session
def delete_event(db, uid):
    '''Delete a specific event identified by its uid. Note that only recorded
    events can be deleted. Events in the buffer for upcoming events are
    regularly replaced anyway and a manual removal could have unpredictable
    effects.

    Use ?hard=true parameter to delete the recorded files on disk as well.

    Returns 204 if the action was successful.
    Returns 404 if event does not exist
    '''
    logger.info('deleting event %s via api', uid)
    events = db.query(RecordedEvent).filter(RecordedEvent.uid == uid)
    if not events.count():
        return make_error_response('No event with specified uid', 404)
    hard_delete = request.args.get('hard', 'false')
    if hard_delete == 'true':
        logger.info('deleting recorded files at %s', events[0].directory())
        shutil.rmtree(events[0].directory())
    events.delete()
    db.commit()
    return make_response('', 204)


@app.route('/api/events/<uid>', methods=['PATCH'])
@requires_auth
@jsonapi_mediatype
@with_session
def modify_event(db, uid):
    '''Modify an event specified by its uid. The modifications for the event
    are expected as JSON with the content type correctly set in the request.

    Note that this method works for recorded events only. Upcoming events part
    of the scheduler cache cannot be modified.
    '''
    try:
        data = request.get_json()['data'][0]
        if data['type'] != 'event' or data['id'] != uid:
            return make_error_response('Invalid data', 400)
        # Check attributes
        for key in data['attributes'].keys():
            if key not in ('status', 'start', 'end'):
                return make_error_response('Invalid data', 400)
        # Check new status
        new_status = data['attributes'].get('status')
        if new_status:
            new_status = new_status.upper().replace(' ', '_')
            data['attributes']['status'] = int(getattr(Status, new_status))
    except Exception:
        return make_error_response('Invalid data', 400)

    event = db.query(RecordedEvent).filter(RecordedEvent.uid == uid).first()
    if not event:
        return make_error_response('No event with specified uid', 404)
    event.start = data['attributes'].get('start', event.start)
    event.end = data['attributes'].get('end', event.end)
    event.status = data['attributes'].get('status', event.status)
    logger.debug('Updating event %s via api', uid)
    db.commit()
    return make_data_response(event.serialize())


@app.route('/api/metrics', methods=['GET'])
@requires_auth
@with_session
def metrics(dbs):
    '''Serve several metrics about the pyCA services and the machine via
    json.'''
    # Get Disk Usage
    # If the capture directory do not exists, use the parent directory.
    directory = config('capture', 'directory')
    if not os.path.exists(directory):
        directory = os.path.abspath(os.path.join(directory, os.pardir))
    total, used, free = shutil.disk_usage(directory)

    # Get Loads
    load_1m, load_5m, load_15m = os.getloadavg()

    # Get Memory
    memory = psutil.virtual_memory()

    # Get Services
    srvs = dbs.query(ServiceStates)
    services = []
    for srv in srvs:
        services.append({
            'name': Service.str(srv.type),
            'status': ServiceStatus.str(srv.status)
        })
    # Get Upstream State
    state = dbs.query(UpstreamState).filter(
        UpstreamState.url == config('server', 'url')).first()
    last_synchronized = state.last_synced.isoformat() if state else None
    return make_response(jsonify(
        {'meta': {
            'services': services,
            'disk_usage_in_bytes': {
                'total': total,
                'used': used,
                'free': free,
            },
            'memory_usage_in_bytes': {
                'total': memory.total,
                'available': memory.available,
                'used': memory.used,
                'free': memory.free,
                'cached': memory.cached,
                'buffers': memory.buffers,
            },
            'load': {
                '1m': load_1m,
                '5m': load_5m,
                '15m': load_15m,
            },
            'upstream': {
                'last_synchronized': last_synchronized,
            }
        }}))



# -*- coding: utf-8 -*-

'''
Tests for database
'''

import unittest
import tempfile
import os

from pyca import db, config


class TestPycaDb(unittest.TestCase):

    dbfile = None

    def setUp(self):
        cfg = './etc/pyca.conf'
        config.update_configuration(cfg)

        self.fd, self.dbfile = tempfile.mkstemp()
        config.config()['agent']['database'] = 'sqlite:///' + self.dbfile

    def tearDown(self):
        os.close(self.fd)
        os.remove(self.dbfile)

    def test_get_session(self):
        assert 'autocommit' in db.get_session().__dict__.keys()

    def test_event_data(self):
        series = u'äöüßÄÖÜ'
        title = u'„xyz“'

        e = db.BaseEvent()
        e.set_data({'series': series, 'title': title})

        # Check data serialization
        data = e.get_data()
        assert data['title'] == title
        assert data['series'] == series

    def test_status(self):
        assert db.Status.str(db.Status.UPCOMING) == 'upcoming'

    def test_event(self):
        e = db.BaseEvent()
        e.uid = 'asd'
        e.start = 123
        e.end = 234
        e.status = db.Status.UPCOMING
        e.set_data({})

        assert str(e) == '<Event(start=123, uid="asd")>'

        e = db.RecordedEvent(e)
        assert e.name() == 'recording-123-asd'
        assert e.status_str() == 'upcoming'
        assert e.serialize()['id'] == 'asd'
        assert e.get_tracks() == []

    def test_servicestate(self):
        s = db.ServiceStates()
        s.type = 0
        s.status = 0
        s = db.ServiceStates(s)

        assert s.type == 0
        assert s.status == 0



# -*- coding: utf-8 -*-
'''
pyCA tests for the agents state handling
'''

import os
import os.path
import tempfile
import unittest

from pyca import agentstate, config, db, utils
from tests.tools import terminate_fn, reload


class TestPycaAgentState(unittest.TestCase):

    def setUp(self):
        utils.http_request = lambda x, y=False: b'xxx'
        self.fd, self.dbfile = tempfile.mkstemp()
        config.config()['agent']['database'] = 'sqlite:///' + self.dbfile
        config.config()['service-capture.admin'] = ['']

        # Mock event
        db.init()

    def tearDown(self):
        os.close(self.fd)
        os.remove(self.dbfile)
        reload(utils)

    def test_run(self):
        agentstate.terminate = terminate_fn(1)
        try:
            agentstate.run()
        except Exception:
            assert False
